from ._FromPredKeepRank import FromPredKeepRank
from ._GenDropRanks import GenDropRanks
from ._GenDropRanksNaive import GenDropRanksNaive

_GenDropRanks_impls = (FromPredKeepRank, GenDropRanks, GenDropRanksNaive)
