"""Tools to create extant populations of hereditary stratigraph columns
as if they had resulted from a given template phylogeny."""

from ..._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub
