from ._HereditaryStratigraphicAssemblage import (
    HereditaryStratigraphicAssemblage,
)
from ._HereditaryStratigraphicAssemblageSpecimen import (
    HereditaryStratigraphicAssemblageSpecimen,
)
from ._HereditaryStratigraphicSpecimen import HereditaryStratigraphicSpecimen

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "HereditaryStratigraphicAssemblage",
    "HereditaryStratigraphicAssemblageSpecimen",
    "HereditaryStratigraphicSpecimen",
]
